import importlib.util
import sys


class LazyImportFinder:
    def __init__(self):
        self._importing = set()  # Track what we're currently importing

    def find_spec(self, fullname, path, target=None):
        # Prevent recursion
        if fullname in self._importing:
            return None

        if not self._should_be_lazy(fullname):
            return None

        self._importing.add(fullname)
        try:
            # Find the real spec using the original meta_path (excluding ourselves)
            spec = None
            for finder in sys.meta_path:
                if finder is self:
                    continue
                if hasattr(finder, 'find_spec'):
                    spec = finder.find_spec(fullname, path, target)
                    if spec is not None:
                        break

            if spec is not None and hasattr(spec.loader, 'exec_module'):
                # Wrap with LazyLoader
                spec.loader = importlib.util.LazyLoader(spec.loader)

            return spec
        finally:
            self._importing.discard(fullname)

    def _should_be_lazy(self, fullname):
        for lazy_module in _lazy_modules:
            if fullname == lazy_module or fullname.startswith(lazy_module + '.'):
                return True
        return False


# Mark which modules should be lazy (base modules)
# would love to lazy load torch and transformers but their import process is too complicated
_lazy_modules = {'bsparse', 'faiss'}
sys.meta_path.insert(0, LazyImportFinder())



import json
from typing import Any, Dict

from ..config import ColllectionConfig
from ..utils import load_singleton
from .abstract import Processor
from .file_random_access_reader import MSMARCOSegOffset, OffsetFile, RandomAccessReader


class ContentProcessor(Processor):
    def __init__(self, collection_config: ColllectionConfig, cache_size=0, cache_ttl=600):
        # always use `id` from the request as the key, this is different from id_field in config
        super().__init__(cache_size, cache_ttl, lambda x: x["id"])

        self.config = collection_config
        self.line_reader = self._load_reader()

        self.content_field = collection_config.content_field

        self.lang_mapping = None
        if collection_config.id_to_lang_mapping is not None:
            self.lang_mapping: Dict[str, str] = load_singleton(collection_config.id_to_lang_mapping)

    def _load_reader(self) -> RandomAccessReader:
        if self.config.offset_source == "offsetfile":
            return OffsetFile(
                self.config.doc_path,
                key=lambda line: json.loads(line)[self.config.id_field],
                offset_fn=self.config.cache_path,
                id_field=self.config.id_field,  # Added for parallel offset map building
            )
        elif self.config.offset_source == "msmarco_seg":
            return MSMARCOSegOffset(self.config.doc_path)

    def __getitem__(self, idx: str):
        doc = json.loads(self.line_reader[idx])
        results = {"text": "\n".join(doc[c] for c in self.content_field)}
        if "title" in doc:
            results["title"] = doc["title"]
        if self.lang_mapping is not None:
            results["language"] = self.lang_mapping.get(idx, "")

        return results

    def __contains__(self, idx: str):
        return idx in self.line_reader

    async def _submit(self, item: Dict[str, Any]) -> Dict[str, str]:
        return (
            self[self.cache_key(item)] if self.cache_key(item) in self else {"error": f"ID {self.cache_key(item)} is not found."}
        )



import logging
import pickle
from abc import ABC
from pathlib import Path
from typing import Any, Dict, List

import aiohttp
from tqdm.auto import tqdm

from .lazy_import import _lazy_modules


logging.basicConfig(
    format="[%(asctime)s][%(levelname)s][%(name)s] %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
    level=logging.INFO,
)

logger = logging.getLogger("search-service")


def pbar(it=None, **kwargs):
    if "dynamic_ncols" not in kwargs:
        kwargs["dynamic_ncols"] = True

    return tqdm(it, **kwargs)


_file_singleton = {}


def load_singleton(fn, load_fn=None):
    global _file_singleton
    fn = str(Path(fn).absolute())
    if fn not in _file_singleton:
        if load_fn is None:
            with open(fn, "rb") as f:
                _file_singleton[fn] = pickle.load(f)
        else:
            _file_singleton[fn] = load_fn(fn)

    return _file_singleton[fn]


def dict_topk(scores: Dict[Any, float], k: int) -> Dict[Any, float]:
    return dict(sorted(scores.items(), key=lambda x: -x[1])[:k])

async def session_request(session: aiohttp.ClientSession, url: str, payload: Dict[str, Any]=None, method="POST"):
    if method == "POST":
        async with session.post(url, json=payload) as response:
            return await response.json()
    elif method == "GET":
        async with session.get(url) as response:
            return await response.json()


def _recursive_subclasses(cls: type):
    for subcls in cls.__subclasses__():
        yield subcls
        yield from _recursive_subclasses(subcls)


class FactoryEnabled(ABC):
    @classmethod
    def load(cls, cls_name: str, **kwargs):
        for subcls in _recursive_subclasses(cls):
            if subcls.__name__ == cls_name or subcls == cls_name:
                return subcls(**kwargs)

        raise TypeError(f"Unsupported subclass {cls_name} for {cls.__name__}")



import asyncio
import json
from pathlib import Path
from typing import Any, Dict, List, Union

import aiohttp

from ..utils import FactoryEnabled, dict_topk, session_request


class Engine(FactoryEnabled):
    def __init__(self, name: str = None, config: Union[str, Path, Dict[str, Any]] = None, **kwargs):
        if config is None:
            config = {}
        elif not isinstance(config, dict):
            config = json.loads(Path(config).read_text())

        self.config: Dict[str, Any] = {**config, **kwargs}

        self.name: str = name or self.config.get("name", None)
        if "index_path" in self.config:
            self.index_path: Path = Path(self.config["index_path"])
        else:
            self.index_path: Path = None

    async def search_batch(self, queries: List[str], limit: Union[int, List[int]] = 20, **kwargs) -> List[Dict[str, float]]:
        raise NotImplementedError

    async def search(self, query: str, limit: int = 20, **kwargs) -> Dict[str, float]:
        return (await self.search_batch([query], limit, **kwargs))[0]

    async def score_batch(
        self, queries: List[str], passages: List[str], candidate_length: List[int] = None, **kwargs
    ) -> List[List[float]]:
        raise NotImplementedError

    async def score(self, query: str, passages: List[str], **kwargs) -> List[float]:
        raise (await self.score_batch([query], passages, [len(passages)]))[0]

    async def decompose_query_batch(self, queries: List[str], limit: List[int] = None, **kwargs) -> List[List[str]]:
        raise NotImplementedError

    async def decompose_query(self, query: str, **kwargs) -> List[str]:
        return (await self.decompose_query_batch([query], **kwargs))[0]

    async def fuse_batch(
        self, queries: List[str], batch_scores: List[List[Dict[str, float]]], **kwargs
    ) -> List[Dict[str, float]]:
        raise NotImplementedError

    async def fuse(self, query: str, scores: List[Dict[str, float]], **kwargs) -> Dict[str, float]:
        return (await self.fuse_batch([query], [scores], **kwargs))[0]

    @property
    def can_search(self) -> bool:
        return self.__class__.search_batch != Engine.search_batch

    @property
    def can_score(self) -> bool:
        return self.__class__.score_batch != Engine.score_batch

    @property
    def can_decompose_query(self) -> bool:
        return self.__class__.decompose_query_batch != Engine.decompose_query_batch

    @property
    def can_fuse(self) -> bool:
        return self.__class__.fuse_batch != Engine.fuse_batch


class Reranker(Engine):
    def __init__(self, name=None, config=None, **kwargs):
        super().__init__(name, config, **kwargs)

        self.upstream = None
        if "upstream_service" in self.config:
            self.upstream = Engine.load(
                self.config["upstream_service"]["engine"], config=self.config["upstream_service"]["config"], **kwargs
            )

        self.text_service = None
        if "text_service" in self.config:
            # Optional, or else you would need to get the document text in other ways
            # e.g. ir_datasets
            assert "endpoint" in self.config["text_service"]
            assert "collection" in self.config["text_service"]
            self.text_service = self.config["text_service"]

        self.rerank_topk_max = int(self.config.get("rerank_topk_max", 100))
        self.rerank_multiplier = float(self.config.get("rerank_multiplier", 5))

    async def get_text(self, docids: Union[str, List[str]]):
        if self.text_service is None:
            raise RuntimeError("No text service provided. Either missing in config or needs to be implemented in subclass.")

        docids = [docids] if isinstance(docids, str) else docids
        async with aiohttp.ClientSession() as session:
            resps = await asyncio.gather(
                *[
                    session_request(
                        session,
                        self.text_service["endpoint"] + "/content",
                        {"collection": self.text_service["collection"], "id": docid},
                    )
                    for docid in set(docids)
                ]
            )
            return {resp["id"]: resp["text"] for resp in resps}

    async def search_batch(self, queries, limit=20, **kwargs) -> List[Dict[str, float]]:
        if self.upstream is None:
            raise RuntimeError(f"Upstream retrieval is not defined, {self.name} only support scoring.")

        if not isinstance(limit, list):
            limit = [limit] * len(queries)
        assert len(limit) == len(queries)

        multiplier = kwargs.get("rerank_multiplier", self.rerank_multiplier)

        upstream_limit = [min(k * multiplier, self.rerank_topk_max) for k in limit]

        upstream_results = await self.upstream.search_batch(queries, limit=upstream_limit, **kwargs)
        candidate_docids = [list(upr.keys()) for upr in upstream_results]

        all_text = await self.get_text(sum(candidate_docids, []))

        candidate_text = [all_text[c] for candidates in candidate_docids for c in candidates]
        candidate_lengths = [len(c) for c in candidate_docids]

        raw_scores = await self.score_batch(queries, candidate_text, candidate_lengths, **kwargs)

        return [
            dict_topk(dict(zip(candidates, qscores)), k) for qscores, candidates, k in zip(raw_scores, candidate_docids, limit)
        ]


class Aggregation:
    def __init__(self, passage_mapping: Dict[str, str]):
        self.mapping = passage_mapping

    def __contains__(self, pid: str) -> bool:
        return pid in self.mapping

    def __getitem__(self, pid: str) -> str:
        return self.mapping[pid]

    def maxp(self, passage_scores: Dict[str, float]) -> Dict[str, float]:
        ret = {}
        for pid, score in passage_scores.items():
            doc_id = self.mapping[pid]
            if doc_id not in ret or ret[doc_id] < score:
                ret[doc_id] = score
        return ret

    @property
    def n_docs(self):
        return len(set(self.mapping.values()))

    def __len__(self):
        return len(self.mapping)


class OptimizedAggregation(Aggregation):
    """Optimized version of Aggregation with pre-computed structures."""

    def __init__(self, passage_mapping: List[str]):
        # Convert list to dict with integer indices for faster lookup
        self.mapping = {str(i): doc_id for i, doc_id in enumerate(passage_mapping)}

        # Pre-compute unique documents for faster statistics
        self._unique_docs = set(self.mapping.values())

        # Create reverse mapping for potential optimizations
        self.reverse_mapping = {}
        for pid, doc_id in self.mapping.items():
            if doc_id not in self.reverse_mapping:
                self.reverse_mapping[doc_id] = []
            self.reverse_mapping[doc_id].append(pid)

    def maxp(self, passage_scores: Dict[str, float]) -> Dict[str, float]:
        """Optimized MaxP using dictionary comprehension."""
        ret = {}
        for pid, score in passage_scores.items():
            if pid in self.mapping:
                doc_id = self.mapping[pid]
                if doc_id not in ret:
                    ret[doc_id] = score
                elif score > ret[doc_id]:
                    ret[doc_id] = score
        return ret

    @property
    def n_docs(self):
        return len(self._unique_docs)



import asyncio
from typing import Any, Dict, List

import aiohttp

from ..processors.registry import ProcessorRegistry
from ..utils import session_request
from .abstract import Engine


class Relay(Engine):
    def __init__(self, name: str = None, config=None, **kwargs):
        super().__init__(name, config, **kwargs)

        assert "service" in self.config

        self.other_kwargs = self.config.get("other_request_kwargs", {})
        # TODO: should support some runtime config like retry and timeout
        # TODO: support list of endpoints for load balancing

    async def _submit_payload(self, service_type, payloads: List[Dict[str, Any]]):
        if "endpoint" in self.config:
            async with aiohttp.ClientSession() as session:
                resps = await asyncio.gather(
                    *[session_request(session, f"{self.config['endpoint']}/{service_type}", load) for load in payloads]
                )
        else:
            assert ProcessorRegistry.has_service(self.config["service"], service_type)
            local_processor = ProcessorRegistry.get(self.config["service"], service_type)
            resps = await asyncio.gather(*[local_processor.submit(load) for load in payloads])

        for resp, payload in zip(resps, payloads):
            assert resp["query"] == payload['query']
        return [
            # for backward compatiblity if the service is using `result` as key
            resp.get("scores", resp.get("result", {})) for resp in resps
        ]


    async def search_batch(self, queries, subsets=None, **kwargs):
        if subsets is None:
            subsets = ["none"] * len(queries)
        assert len(subsets) == len(queries)

        for key in kwargs:
            if isinstance(kwargs[key], list):
                assert len(kwargs[key]) == len(queries)
            else:
                kwargs[key] = [kwargs[key]] * len(queries)

        return await self._submit_payload("search", [
            {
                "query": queries[i],
                "service": self.config["service"],
                "subset": subsets[i],
                **self.other_kwargs,
                **{k: kwargs[k][i] for k in kwargs},
            }
            for i in range(len(queries))
        ])


    async def score_batch(self, queries, passages, candidate_length = None, **kwargs):
        if candidate_length is None:
            candidate_length = [len(passages)]
        assert len(candidate_length) == len(queries)
        assert sum(candidate_length) == len(passages)

        payloads = []
        start = 0
        for query, l in zip(queries, candidate_length):
            payloads.append({
                "query": query,
                "service": self.config["service"],
                "passages": passages[start: start+l]
            })
            start = start + 1

        return await self._submit_payload("score", payloads)



from .abstract import BatchProcessor, LRUCache, Processor
from .content_processors import ContentProcessor
from .file_random_access_reader import OffsetFile
from .query_processors import AsyncQueryProcessor, BatchQueryProcessor
from .registry import ProcessorRegistry, auto_register
from .score_processors import AsyncPairwiseScoreProcessor, BatchPairwiseScoreProcessor



from dataclasses import dataclass
from typing import List, Optional, Union

from lark import Lark, Transformer


PIPELINE_GRAMMAR = r"""
    ?start: seq

    seq: stage (">>" stage)*

    stage: (parallel_seq | system_call)

    system_call: NAME alias? ("%" NUMBER)?

    alias: ("[" NAME "]")

    seq_list: seq ("," seq)*

    parallel_seq: system_call? "{" seq_list "}" system_call 

    NAME: /[A-Za-z][A-Za-z0-9_\-]*/
    NUMBER: /[0-9]+/

    %import common.WS
    %ignore WS
"""

# TODO: adding `:collection` into grammar to specify where to get the content


# Data classes to represent the AST nodes
@dataclass
class SystemCall:
    name: str
    alias: Optional[str] = None
    limit: Optional[int] = None
    role: Optional[str] = "search"

    def __post_init__(self):
        if self.alias is None:
            self.alias = self.name

    @property
    def all_calls(self):
        return set([self])

    def __hash__(self):
        return (self.name, self.alias, self.limit).__hash__()

    def as_role(self, role: str):
        return SystemCall(self.name, self.alias, self.limit, role)


@dataclass
class CallSequence:
    stages: List[Union[SystemCall, "ParallelCallSequences"]]

    def __post_init__(self):
        if len(self.stages) > 1:
            self.stages = [self.stages[0]] + [s.as_role("rerank") for s in self.stages[1:]]

    @property
    def all_calls(self):
        return set.union(*[s.all_calls for s in self.stages])

    def as_role(self, role: str):
        assert role in ["search", "rerank"]
        return CallSequence([self.stages[0].as_role(role), *self.stages[1:]])


@dataclass
class ParallelCallSequences:
    sequences: List[CallSequence]
    merger: SystemCall
    expander: Optional[SystemCall] = None

    def __post_init__(self):
        if self.expander is not None:
            self.expander.role = "expander"
        # else:
        #     self.sequences = [ s.as_role('rerank') for s in self.sequences ]
        self.merger.role = "merger"

    @property
    def all_calls(self):
        return set.union(
            set() if self.expander is None else self.expander.all_calls,
            self.merger.all_calls,
            *[s.all_calls for s in self.sequences],
        )

    def as_role(self, role: str):
        if self.expander is not None:
            # TODO: better handle this but we shouldn't support reranking existing
            # ranked list with expanded queries
            return self
        else:
            assert role in ["search", "rerank"]
            return ParallelCallSequences(sequences=[s.as_role(role) for s in self.sequences], merger=self.merger)


PipelineComponent = Union[SystemCall, CallSequence, ParallelCallSequences]


# Transformer to convert parse tree to AST
class PipelineTransformer(Transformer):
    def seq(self, stages):
        if len(stages) == 1:
            return stages[0].as_role("search")
        return CallSequence(stages=stages)

    def system_call(self, tokens: List[str]):
        name, alias, limit = tokens[0], None, None
        if len(tokens) == 3:
            name, alias, limit = tokens
        elif len(tokens) == 2:
            if tokens[1].isdigit():
                name, limit = tokens
            else:
                name, alias = tokens

        return SystemCall(name=str(name), alias=alias, limit=int(limit) if limit is not None else None)

    def alias(self, tokens):
        return str(tokens[0])

    def parallel_seq(self, tokens):
        if len(tokens) == 3:
            return ParallelCallSequences(expander=tokens[0], sequences=tokens[1], merger=tokens[2])
        return ParallelCallSequences(sequences=tokens[0], merger=tokens[1])

    def stage(self, tokens):
        return tokens[0]

    def seq_list(self, tokens):
        return tokens


parser = Lark(PIPELINE_GRAMMAR, parser="lalr", transformer=PipelineTransformer())



from typing import List, Tuple, Union

from tqdm.auto import tqdm

from ..utils import logger
from .abstract import Reranker


try:
    import torch
    from transformers import AutoTokenizer, BatchEncoding, PreTrainedTokenizer, T5ForConditionalGeneration
except ImportError:
    logger.warning("Failed to import torch and transformers for T5 Rerankers")

DecodedOutput = Union["torch.Tensor", Tuple["torch.Tensor", "torch.Tensor"]]

# copied from PyGaggle repo
# https://github.com/castorini/pygaggle/blob/master/pygaggle/rerank/transformer.py
prediction_tokens = {
    "castorini/monot5-base-msmarco": ["▁false", "▁true"],
    "castorini/monot5-base-msmarco-10k": ["▁false", "▁true"],
    "castorini/monot5-large-msmarco": ["▁false", "▁true"],
    "castorini/monot5-large-msmarco-10k": ["▁false", "▁true"],
    "castorini/monot5-base-med-msmarco": ["▁false", "▁true"],
    "castorini/monot5-3b-med-msmarco": ["▁false", "▁true"],
    "castorini/monot5-3b-msmarco-10k": ["▁false", "▁true"],
    # 'unicamp-dl/mt5-13b-mmarco-100k':            ['▁false', '▁true'], # bug
    "unicamp-dl/mt5-13b-mmarco-100k": ["▁", "▁true"],
    "unicamp-dl/mt5-base-en-msmarco": ["▁no", "▁yes"],
    "unicamp-dl/ptt5-base-pt-msmarco-10k-v2": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-pt-msmarco-100k-v2": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-en-pt-msmarco-100k-v2": ["▁não", "▁sim"],
    "unicamp-dl/mt5-base-en-pt-msmarco-v2": ["▁no", "▁yes"],
    "unicamp-dl/mt5-base-mmarco-v2": ["▁no", "▁yes"],
    "unicamp-dl/mt5-base-en-pt-msmarco-v1": ["▁no", "▁yes"],
    "unicamp-dl/mt5-base-mmarco-v1": ["▁no", "▁yes"],
    "unicamp-dl/ptt5-base-pt-msmarco-10k-v1": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-pt-msmarco-100k-v1": ["▁não", "▁sim"],
    "unicamp-dl/ptt5-base-en-pt-msmarco-10k-v1": ["▁não", "▁sim"],
    "unicamp-dl/mt5-3B-mmarco-en-pt": ["▁", "▁true"],
}


def _get_prediction_tokens(
    pretrained_model_name_or_path: str, tokenizer: "PreTrainedTokenizer", token_false=None, token_true=None
):
    if not (token_false and token_true):
        if pretrained_model_name_or_path in prediction_tokens:
            token_false, token_true = prediction_tokens[pretrained_model_name_or_path]
            token_false_id = tokenizer.get_vocab()[token_false]
            token_true_id = tokenizer.get_vocab()[token_true]
            return token_false_id, token_true_id
        else:
            raise Exception(
                f"We don't know the indexes for the non-relevant/relevant tokens for\
                    the checkpoint {pretrained_model_name_or_path} and you did not provide any."
            )
    else:
        token_false_id = tokenizer.get_vocab()[token_false]
        token_true_id = tokenizer.get_vocab()[token_true]
        return token_false_id, token_true_id


class MT5Reranker(Reranker):
    prompt = "Query: {query} Document: {document} Relevant:"

    def __init__(self, name=None, config=None, **kwargs):
        super().__init__(name, config, **kwargs)

        assert "model_name_or_path" in self.config
        self.q_max_length = int(self.config.get("q_max_length", 180))
        self.d_max_length = int(self.config.get("d_max_length", 512))
        self.batch_size = int(self.config.get("batch_size", 32))

        self.model = T5ForConditionalGeneration.from_pretrained(self.config["model_name_or_path"])
        self.tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(self.config["model_name_or_path"], use_fast=True)
        self.model.eval()

        if self.config.get("use_gpu", True):
            self.model = self.model.bfloat16().cuda()

        self.token_false_id, self.token_true_id = _get_prediction_tokens(self.config["model_name_or_path"], self.tokenizer)

    def tokenize_pairs(self, batch: List[Tuple[str, str]]) -> "BatchEncoding":
        return self.tokenizer(
            [self.prompt.format(query=q, document=d) for q, d in batch],
            return_tensors="pt",
            return_attention_mask=True,
            padding="max_length",
            truncation=True,
            max_length=self.q_max_length + self.d_max_length + 9,
        )

    # copied and modified from PyGaggle repo
    # https://github.com/castorini/pygaggle/blob/master/pygaggle/model/decode.py
    def forward(
        self,
        input_ids: "torch.Tensor",
        attention_mask: "torch.Tensor" = None,
    ) -> DecodedOutput:
        encoder_outputs = self.model.get_encoder()(input_ids, attention_mask=attention_mask)

        with torch.no_grad():
            decode_ids = torch.full(
                (input_ids.size(0), 1), self.model.config.decoder_start_token_id, dtype=torch.long, device=input_ids.device
            )

            model_inputs = self.model.prepare_inputs_for_generation(
                decode_ids, encoder_outputs=encoder_outputs, past_key_values=None, attention_mask=attention_mask, use_cache=True
            )
            outputs = self.model(**model_inputs)  # (batch_size, cur_len, vocab_size)
            true_false_logits = outputs[0][:, -1, [self.token_false_id, self.token_true_id]]  # (batch_size, vocab_size)

            return torch.nn.functional.log_softmax(true_false_logits, dim=1)[:, 1].contiguous()

    async def score(self, queries, passages, candidate_length=None, with_progress=False, **kwargs):
        if candidate_length is None:
            if len(queries) == len(passages):  # pairwise
                candidate_length = [1] * len(queries)
            if len(passages) % len(queries) == 0:  # assume same number of candidates
                candidate_length = [len(passages) // len(queries)] * len(queries)
            else:
                raise ValueError("mT5 Reranker does not support all pair scoring.")

        assert len(candidate_length) == len(queries)
        assert sum(candidate_length) == len(passages)

        expanded_queries = sum([[query] * l for query, l in zip(queries, candidate_length)], [])
        # qidx = sum([[qid] * l for qid, l in zip(range(len(queries)), candidate_length)], [])

        all_scores = (
            torch.concat(
                [
                    self.forward(
                        **self.tokenize_pairs(
                            list(zip(expanded_queries[i : i + self.batch_size], passages[i : i + self.batch_size]))
                        ).to(self.model.device)
                    ).cpu()
                    for i in tqdm(range(0, len(passages), self.batch_size), disable=not with_progress, dynamic_ncols=True)
                ]
            )
            .flatten()
            .tolist()
        )

        offsets = [0] + [sum(candidate_length[: i + 1]) for i in range(len(candidate_length))]

        return [all_scores[sidx:eidx] for sidx, eidx in zip(offsets[:-1], offsets[1:])]



import time
from typing import Any, Dict, List, Union

from ..models.abstract import Engine
from .abstract import Processor


_callable_service_types = {"search", "score", "decompose_query", "fuse"}
_uncallable_service_types = {"content"}

# TODO: why am I creating this mess...
_output_keys = {"search": "scores", "score": "scores", "decompose_query": "queries", "fuse": "scores"}


class _ProcessorRegistry:
    def __init__(self):
        # TODO: make collection name part of the namespace
        self.all_services: Dict[str, Dict[str, Processor]] = {}

    @property
    def valid_service_types(self):
        return _callable_service_types | _uncallable_service_types

    def __getitem__(self, name: str):
        return self.all_services[name]

    def register(self, name: str, service_type: str, processor: Processor):
        assert service_type in self.valid_service_types, f"Invalid service type `{service_type}`."
        assert name not in self.all_services or service_type not in self.all_services[name], (
            f"Service type `{service_type}` of name `{name}` already exists."
        )

        if name not in self.all_services:
            self.all_services[name] = {}
        self.all_services[name][service_type] = processor

    def has_service(self, name: str, service_type: str):
        return name in self.all_services and service_type in self.all_services[name]

    def get(self, name: str, service_type: str):
        if self.has_service(name, service_type):
            return self.all_services[name][service_type]

    def get_all_services(self):
        services = {t: [] for t in self.valid_service_types}
        for name, processors in self.all_services.items():
            for s in processors:
                services[s].append(name)
        return services


class DummyProcessor(Processor):
    def __init__(self, engine: Engine, method: str):
        super().__init__(cache_size=0)
        assert hasattr(engine, method)
        self.service = getattr(engine, method)
        self.result_key = _output_keys[method]

    async def submit(self, item: Dict[str, Any]) -> Dict[str, Any]:
        return {self.result_key: (await self.service(**item)), "processed": True, "timestamp": time.time()}


# singleton
ProcessorRegistry = _ProcessorRegistry()


def auto_register(methods: Union[str, List[str]], **default_init_kwargs):
    if isinstance(methods, str):
        methods = [methods]

    def engine_dec(engine_cls: type[Engine]):
        assert issubclass(engine_cls, Engine)
        assert all(getattr(engine_cls, f"can_{method}") for method in methods)
        # register each
        for method in methods:
            ProcessorRegistry.register(
                engine_cls.__name__, method, processor=DummyProcessor(engine_cls(**default_init_kwargs), method)
            )

        return engine_cls

    return engine_dec



import time
from typing import Any, Dict, List

from ..models import Engine
from .abstract import BatchProcessor, Processor


class AsyncPairwiseScoreProcessor(Processor):
    """Processor that serve every request independently through async calls.
    Should be used when all the engine does is issuing async calls
    """

    def __init__(self, engine: Engine, cache_size=1024, cache_ttl=600, cache_key=None, **kwargs):
        super().__init__(cache_size, cache_ttl, cache_key)
        self.engine = engine

    async def _submit(self, item: Dict[str, Any]):
        """
        Process one single item
        """

        query = item.pop("query")

        ranking = (await self.engine.score_batch([query], passages=item.pop("passages")))[0]

        # Process each item in the batch
        return {"query": query, "scores": ranking, "service": self.engine.name, "processed": True, "timestamp": time.time()}


class BatchPairwiseScoreProcessor(BatchProcessor):
    def __init__(self, engine: Engine, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine

    async def _process_batch(self, batch: List[Dict]) -> List[Dict]:
        """
        Process a batch of items.
        Override this method for specific batch processing logic.
        """
        # Simulate processing time (e.g., model inference)
        # await asyncio.sleep(0.5)
        queries = [item.get("query", "") for item in batch]
        passages = [item.get("passages", []) for item in batch]

        batch_scores = await self.engine.score_batch(queries, sum(passages, []), list(map(len, passages)))

        # Process each item in the batch
        results = []
        for item, scores in zip(batch, batch_scores):
            result = {
                "meta": {"n_passages": len(item["passages"])},
                "query": item["query"],
                "scores": scores,
                "service": self.engine.name,
                "processed": True,
                "timestamp": time.time(),
            }
            results.append(result)

        return results



import argparse
import asyncio

from hypercorn.asyncio import serve
from hypercorn.config import Config
from quart import Quart, jsonify, request

from .config.load import load_config
from .pipeline import SearchPipeline
from .processors import ProcessorRegistry
from .utils import logger


app = Quart(__name__)
config = None


@app.before_serving
async def startup():
    """Initialize resources before the server starts."""
    global config
    await load_config(config)


# TODO: standardize the API format with pydantic


@app.route("/search", methods=["POST"])
@app.route("/query", methods=["POST"])  # deprecated
async def process_query():
    """API endpoint for processing requests."""

    try:
        data = await request.get_json()
        if not data:
            return jsonify({"error": "No data provided"}), 400

        service = data.pop("service")
        if not ProcessorRegistry.has_service(service, "search"):
            return jsonify({"error": "Unsupported service"}), 400

        result = await ProcessorRegistry.get(service, "search").submit(data)
        return jsonify(result)

    except Exception as e:
        logger.error(f"Error processing request: {e}")
        return jsonify({"error": str(e)}), 500


@app.route("/score", methods=["POST"])
async def process_scoring():
    """API endpoint for processing requests."""

    try:
        data = await request.get_json()
        if not data:
            return jsonify({"error": "No data provided"}), 400

        service = data.pop("service")
        if not ProcessorRegistry.has_service(service, "score"):
            return jsonify({"error": "Unsupported service"}), 400

        result = await ProcessorRegistry.get(service, "score").submit(data)
        return jsonify(result)

    except Exception as e:
        logger.error(f"Error processing request: {e}")
        return jsonify({"error": str(e)}), 500


@app.route("/content", methods=["POST"])
async def process_get_content():
    try:
        data = await request.get_json()
        if not data or "id" not in data:
            return jsonify({"error": "No id provided"}), 400

        if not ProcessorRegistry.has_service(data["collection"], "content"):
            return jsonify({"error": "Unsupported collection"}), 400

        result = await ProcessorRegistry.get(data["collection"], "content").submit(data)
        return jsonify({**data, **result}), 400 if "error" in result else 200

    except Exception as e:
        logger.error(f"Error processing request: {e}")
        return jsonify({"error": str(e)}), 500


@app.route("/pipeline", methods=["POST"])
async def process_pipeline():
    try:
        data = await request.get_json()
        if not data:
            return jsonify({"error": "No data provided"}), 400
        # required fields
        for field in ["pipeline", "collection", "query"]:
            if field not in data:
                return jsonify({"error": f"No {field} provided"}), 400

        pipeline = SearchPipeline.from_string(data["pipeline"], data["collection"], runtime_kwargs=data.get("runtime_kwargs", {}))
        result = await pipeline.run(data["query"])
        return jsonify({**data, **result}), 200

    except Exception as e:
        logger.error(f"Error processing request: {e}")
        return jsonify({"error": str(e)}), 500


@app.route("/ping", methods=["GET"])
async def health_check():
    """Simple health check endpoint."""
    return jsonify({"status": "pong"})


@app.route("/avail", methods=["GET"])
async def get_avail_service():
    return jsonify(ProcessorRegistry.get_all_services())


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("config", type=str)
    parser.add_argument("--port", type=int, default=5000)
    parser.add_argument("--host", type=str, default="0.0.0.0")
    parser.add_argument("--cache_dir", type=str, default="./.cache")

    args = parser.parse_args()

    global config
    config = args.config

    # app.run(host=args.host, port=args.port, use_reloader=False)
    hypercorn_config = Config()
    hypercorn_config.bind = [f"{args.host}:{args.port}"]
    asyncio.run(serve(app, hypercorn_config))


if __name__ == "__main__":
    main()



import importlib.metadata
import importlib.util
import sys
from pathlib import Path
from types import ModuleType

from . import logger


def _explore_modules(module: ModuleType, scope=None):
    if scope is None:
        scope = globals()

    names = getattr(module, "__all__", [n for n in dir(module) if not n.startswith("_")])
    for name in names:
        scope[name] = getattr(module, name)


def load_all_extensions(user_specified_files=None, scope=None, use_entry_points=True):
    """
    Load routir extensions using both naming convention and entry points.

    Args:
        user_specified_files: List of files to import
        scope: Namespace to import into (defaults to globals())
        use_entry_points: If True, prefer entry points; if False, use prefix only
    """
    if scope is None:
        scope = globals()

    if user_specified_files is None:
        user_specified_files = []

    loaded = []

    for fn in user_specified_files:
        try:
            fn = Path(fn).resolve()

            if not fn.is_file():
                raise FileNotFoundError(f"{fn} is not a valid file")

            logger.warning(f"Importing from `{fn}`: should only do so if you trust the source. ")
            spec = importlib.util.spec_from_file_location(fn.stem, str(fn))
            module = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = module
            spec.loader.exec_module(module)

            _explore_modules(module, scope)

            loaded.append(f"{fn.stem} (script)")

        except Exception as e:
            logger.warning(f"Failed to load from file {fn}: {e}")

    if use_entry_points:
        # Try entry points first
        entry_points = importlib.metadata.entry_points()

        if hasattr(entry_points, "select"):
            routir_entries = entry_points.select(group="routir.extensions")
        else:
            routir_entries = entry_points.get("routir.extensions", [])

        for entry_point in routir_entries:
            try:
                plugin = entry_point.load()
                _explore_modules(plugin, scope)
                loaded.append(f"{entry_point.name} (entry point)")
            except Exception as e:
                logger.warning(f"Failed to load {entry_point.name}: {e}")

    # Also scan for routir_* packages
    for dist in importlib.metadata.distributions():
        package_name = dist.metadata["Name"]

        if package_name.startswith("routir_"):
            try:
                module_name = package_name.replace("-", "_")

                # Skip if already loaded via entry point or files
                if any(module_name in item for item in loaded):
                    continue

                module = importlib.import_module(module_name)
                _explore_modules(module, scope)

                loaded.append(f"{package_name} (prefix)")
            except Exception as e:
                logger.warning(f"Failed to load {package_name}: {e}")

    if len(loaded) > 0:
        logger.warning("Loaded the following extensions: " + "\n    ".join([""] + loaded))



import asyncio
from typing import Any, Dict, List

from ..processors.registry import ProcessorRegistry
from ..utils import dict_topk
from .parser import CallSequence, ParallelCallSequences, PipelineComponent, SystemCall, parser


# TODO: probably should unify it...
_role_to_service = {"search": "search", "rerank": "score", "expander": "decompose_query", "merger": "fuse"}


class SearchPipeline:
    def __init__(
        self, pipeline: PipelineComponent, collection: str, runtime_kwargs: Dict[str, Dict[str, Any]] = None, verify: bool = True
    ):
        self.pipeline = pipeline
        self.collection = collection
        self.runtime_kwargs = runtime_kwargs or {}
        self.doc_content_cache = {}

        if verify:
            self.verify()
        alias_not_found = set(self.runtime_kwargs.keys()) - set([c.alias for c in self.pipeline.all_calls])
        assert len(alias_not_found) == 0, f"Cannot find alias {alias_not_found}"

    def verify(self):
        assert ProcessorRegistry.has_service(self.collection, "content")
        for call in self.pipeline.all_calls:
            assert ProcessorRegistry.has_service(call.name, _role_to_service[call.role]), (
                f"Cannot find a {_role_to_service[call.role]} service under `{call.name}`"
            )

    @classmethod
    def from_string(
        cls, pipeline_string: str, collection: str, runtime_kwargs: Dict[str, Dict[str, Any]] = None, verify: bool = True
    ) -> "SearchPipeline":
        return cls(parser.parse(pipeline_string), collection, runtime_kwargs, verify)

    async def get_doc_content(self, doc_id: str):
        if doc_id not in self.doc_content_cache:
            ret = await ProcessorRegistry[self.collection]["content"].submit({"id": doc_id})
            self.doc_content_cache[doc_id] = ret["text"]
        return self.doc_content_cache[doc_id]

    async def run(
        self,
        query: str,
        last_output: Dict[str, Any] = None,
        current_node: PipelineComponent = None,
        scratch: Dict[str, Dict[str, Any]] = None,
    ) -> List[Dict[str, float]]:
        current_node = current_node or self.pipeline
        last_output = last_output or {}
        scratch = scratch or {}

        if isinstance(current_node, CallSequence):
            for stage in current_node.stages:
                last_output = await self.run(query, last_output, stage, scratch)
            return last_output

        if isinstance(current_node, ParallelCallSequences):
            expanded_queries = [query]
            if current_node.expander is not None:
                last_output = await self.run(query, last_output, current_node.expander, scratch)
                expanded_queries = last_output["queries"]

            concurrent_run_outputs = await asyncio.gather(
                *[self.run(q, last_output, seq, scratch) for seq in current_node.sequences for q in expanded_queries]
            )

            return await self.run(
                query, {"scores": [o["scores"] for o in concurrent_run_outputs], **last_output}, current_node.merger, scratch
            )

        assert isinstance(current_node, SystemCall)

        payload = {"query": query, **last_output, **scratch, **self.runtime_kwargs.get(current_node.alias, {})}
        if current_node.limit is not None:
            payload["limit"] = current_node.limit

        processor = ProcessorRegistry[current_node.name][_role_to_service[current_node.role]]

        if current_node.role == "search":
            ret = await processor.submit(payload)
            assert "scores" in ret and isinstance(ret["scores"], dict)

        if current_node.role == "merger":
            ret = await processor.submit(payload)
            assert "scores" in ret and isinstance(ret["scores"], dict)

        if current_node.role == "rerank":
            docid_to_rerank: List[str] = list(last_output["scores"].keys())
            doc_text_list = await asyncio.gather(*[self.get_doc_content(d) for d in docid_to_rerank])
            payload["passages"] = doc_text_list
            ret = await processor.submit(payload)
            assert "scores" in ret and isinstance(ret["scores"], list)
            ret["scores"] = dict(zip(docid_to_rerank, ret["scores"]))

        if current_node.role == "expander":
            ret = await processor.submit(payload)
            assert "queries" in ret and isinstance(ret["queries"], list)

        # apply limit here just to safe
        if current_node.limit is not None:
            if "scores" in ret:
                ret["scores"] = dict_topk(ret["scores"], current_node.limit)
            elif "queries" in ret:
                ret["queries"] = ret["queries"][: current_node.limit]

        scratch[(current_node.alias, current_node.role)] = ret
        return ret



import asyncio
from typing import Callable, Dict, List

from ..processors.registry import auto_register
from ..utils import dict_topk
from .abstract import Engine


def _rrf(score_list: List[Dict[str, float]], smoothing_k=0) -> Dict[str, float]:
    fused = {}
    for scores in score_list:
        for i, (docid, s) in enumerate(sorted(scores.items(), key=lambda x: -x[1])):
            if docid not in fused:
                fused[docid] = 0
            fused[docid] += 1 / (i + 1 + smoothing_k)

    return fused


def _score_fusion(score_list: List[Dict[str, float]]) -> Dict[str, float]:
    fused = {}
    for scores in score_list:
        for docid, s in scores.items():
            if docid not in fused:
                fused[docid] = 0
            fused[docid] += s

    return fused


_fusion_functions: Dict[str, Callable] = {
    "RRF": _rrf,
    "score": _score_fusion,
}


class Fusion(Engine):
    # NOTE: this might be deprecate at some point but it is still useful when specifying a specific
    # process using config

    def __init__(self, name=None, config=None, **kwargs):
        super().__init__(name, config, **kwargs)

        self.upstream: List[Engine] = []
        if not isinstance(self.config["upstream_service"], list):
            self.config["upstream_service"] = [self.config["upstream_service"]]
        for c in self.config["upstream_service"]:
            self.upstream.append(Engine.load(c["engine"], config=c["config"], **kwargs))

        assert len(self.upstream) > 0, "Need to have at least one upstream service"

        self.fusion_function = _fusion_functions[self.config.get("fusion_method", "RRF")]
        self.fusion_args = self.config.get("fusion_args", {})

    async def search_batch(self, queries, limit=20, **kwargs) -> List[Dict[str, float]]:
        if not isinstance(limit, list):
            limit = [limit] * len(queries)
        assert len(limit) == len(queries)

        upstream_results = await asyncio.gather(
            *[service.search_batch(queries, limit=limit, **kwargs) for service in self.upstream]
        )

        return [
            dict_topk(self.fusion_function([results[qid] for results in upstream_results], **self.fusion_args), k)
            for qid, k in enumerate(limit)
        ]


@auto_register("fuse")
class RRF(Engine):
    async def fuse_batch(self, queries, batch_rankings, **kwargs):
        return [_rrf(rankings, int(kwargs.get("rrf_k", 0))) for rankings in batch_rankings]


@auto_register("fuse")
class ScoreFusion(Engine):
    async def fuse_batch(self, queries, batch_rankings, **kwargs):
        return [_score_fusion(rankings) for rankings in batch_rankings]



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List

from ...utils import FactoryEnabled


def _trim_indent_in_block(block):
    # only works for spaces, not tabs
    first_line = block.strip("\n").split("\n")[0]
    indent = len(first_line) - len(first_line.lstrip())
    return "\n".join([line[indent:] if len(line) > indent else line for line in block.strip("\n").split("\n")]).strip()


def _combine_passages(passages):
    return "\n\n".join(f"[{i + 1}] {text}" for i, text in enumerate(passages))


@dataclass
class PromptBuilder(FactoryEnabled):
    @abstractmethod
    def __call__(self, query: str, candidates: List[str]) -> List[str]:
        raise NotImplementedError


@dataclass
class ResponseParser(FactoryEnabled):
    @abstractmethod
    def __call__(self, responses: List[str]) -> Dict[int, float]:
        raise NotImplementedError


@dataclass
class RankKPromptBuilder(PromptBuilder):
    prompt = _trim_indent_in_block("""
        Determine a ranking of the passages based on how relevant they are to the query. 
        If the query is a question, how relevant a passage is depends on how well it answers the question. 
        If not, try analyze the intent of the query and assess how well each passage satisfy the intent. 
        The query may have typos and passages may contain contradicting information. 
        However, we do not get into fact-checking. We just rank the passages based on they relevancy to the query. 

        Sort them from the most relevant to the least. 
        Answer with the passage number using a format of `[3] > [2] > [4] = [1] > [5]`. 
        Ties are acceptable if they are equally relevant. 
        I need you to be accurate but overthinking it is unnecessary.
        Output only the ordering without any other text.

        Query: {query}

        {docs}
    """)

    def __call__(self, query, candidates):
        return [self.prompt.format(query=query, docs=_combine_passages(candidates))]


@dataclass
class NumberListParser(ResponseParser):
    def __call__(self, responses: List[str], pids: List[Any]):
        assert len(responses) == 1

        rr = responses[0].strip().split("\n")[-1].split(">")
        ranking = []
        for i, p in enumerate(rr):
            for pidxstring in p.split("="):
                pdix = pidxstring.strip().replace("[", "").replace("]", "")

                try:
                    pid = pids[int(pdix) - 1]
                except KeyboardInterrupt:
                    pass
                except:
                    continue

                if pid in ranking:
                    return ranking
                ranking.append(pid)

        return ranking


"[7] > [3] > [8] > [18] > [12] = [13] = [14] = [10] > [19] > [5] > [20] > [1] = [4] = [6] = [9] = [11] = [15] = [16] = [17]"



import asyncio
import time
import uuid
from typing import Any, Dict, List

from ..utils import FactoryEnabled, logger
from .cache import Cache, LRUCache, RedisCache


class Processor(FactoryEnabled):
    def __init__(self, cache_size=1024, cache_ttl=600, cache_key=None, redis_url: str = None, redis_kwargs: Dict[str, Any] = {}):
        self.cache: Cache = None
        if cache_size > 0 and redis_url is None:
            self.cache = LRUCache(cache_size, cache_ttl)
        elif cache_size > 0 and redis_url is not None:
            self.cache = RedisCache(cache_size, cache_ttl, redis_url, **redis_kwargs)
        self.cache_key = cache_key
        if self.cache_key is None:
            self.cache_key = lambda x: (x.get("service", "default"), x["query"], x.get("limit", "none"), x.get("subset", "none"))

    async def start(self):
        pass

    async def submit(self, item: Any) -> Dict[str, Any]:
        # check cache
        if self.cache is not None:
            cache_key = self.cache_key(item)
            cached = await self.cache.get(cache_key)

            if cached is not None:
                return cached

        result = await self._submit(item)

        # cache the result
        if self.cache is not None and "error" not in result:
            # eventually consistent; no need to wait for cache write before returning
            asyncio.create_task(self.cache.put(cache_key, {**result, "cached": True}))

        return {**result, "cached": False}

    async def _submit(self, item: Any) -> Dict[str, Any]:
        raise NotImplementedError


# Mostly thank to Claude
class BatchProcessor(Processor):
    def __init__(self, batch_size=32, max_wait_time=0.1, cache_size=1024, cache_ttl=600, cache_key=None, **kwargs):
        """
        Simple dynamic batch processor for batching similar requests.

        Args:
            batch_size: Maximum number of items in a batch
            max_wait_time: Maximum time to wait before processing a partial batch
        """
        super().__init__(cache_size, cache_ttl, cache_key)

        self.batch_size = batch_size
        self.max_wait_time = max_wait_time
        # All these will be initialized in start()
        self.queue = None
        self.results: Dict[str, Any] = None
        self.result_events: Dict[str, asyncio.Event] = None
        self.worker_task = None
        self._started = False

    async def start(self):
        """Initialize and start the batch processor."""
        if self._started:
            return

        # Initialize all attributes with the current event loop
        self.queue = asyncio.Queue()
        self.results = {}
        self.result_events = {}

        # Start the worker task
        self.worker_task = asyncio.create_task(self._worker())
        self._started = True
        logger.info("Batch processor started")

    async def _submit(self, item):
        """Submit an item for processing and wait for the result."""
        if not self._started:
            await self.start()

        # Create a unique ID for this request
        item_id = str(uuid.uuid4())

        # Create an event for this request
        event = asyncio.Event()
        self.result_events[item_id] = event

        # Add the item to the queue
        await self.queue.put((item_id, item))
        logger.debug(f"Item {item_id} added to queue")

        # Wait for the result
        await event.wait()

        # Get and return the result
        result = self.results.pop(item_id)
        del self.result_events[item_id]

        return result

    async def _worker(self):
        """Worker that collects items into batches and processes them."""
        logger.info("Worker started")

        while True:
            try:
                batch = []
                batch_ids = []

                # Get the first item (blocks until an item is available)
                try:
                    first_id, first_item = await self.queue.get()
                    batch.append(first_item)
                    batch_ids.append(first_id)
                    logger.debug(f"First item {first_id} received")
                except Exception as e:
                    logger.error(f"Error getting first item: {e}")
                    await asyncio.sleep(0.1)
                    continue

                # Try to fill the batch with more items
                batch_start_time = time.time()
                while len(batch) < self.batch_size and time.time() - batch_start_time < self.max_wait_time:
                    try:
                        # Calculate remaining time
                        remaining_time = max(0.01, self.max_wait_time - (time.time() - batch_start_time))

                        # Try to get another item with timeout
                        item_id, item = await asyncio.wait_for(self.queue.get(), timeout=remaining_time)
                        batch.append(item)
                        batch_ids.append(item_id)
                        logger.debug(f"Additional item {item_id} added to batch")
                    except asyncio.TimeoutError:
                        # No more items available within timeout
                        logger.debug("Timeout waiting for more items")
                        break
                    except Exception as e:
                        logger.error(f"Error collecting batch: {e}")
                        break

                # Process the batch
                batch_size = len(batch)
                logger.info(f"Processing batch of size {batch_size}, {self.queue.qsize()} pending")

                # TODO: should try to dedup the batch based on queries etc

                try:
                    # This is where you'd do your actual batch processing
                    # (e.g., model inference, database operations, etc.)
                    batch_results = await self._process_batch(batch)

                    # Distribute results to waiting clients
                    for i, item_id in enumerate(batch_ids):
                        if i < len(batch_results):
                            self.results[item_id] = batch_results[i]
                        else:
                            # Handle case of mismatched results
                            self.results[item_id] = {"error": "Processing error: missing result"}

                        # Signal that the result is ready
                        if item_id in self.result_events:
                            self.result_events[item_id].set()
                except Exception as e:
                    import traceback

                    logger.error(f"Error processing batch: {traceback.format_exc()}")
                    # Return error to all waiting requests
                    for item_id in batch_ids:
                        self.results[item_id] = {"error": f"Processing error: {str(e)}"}
                        if item_id in self.result_events:
                            self.result_events[item_id].set()

                # Mark tasks as done in the queue
                for _ in range(batch_size):
                    self.queue.task_done()

            except Exception as e:
                logger.error(f"Unexpected error in worker: {e}")
                # Small delay to prevent CPU spinning in case of persistent errors
                await asyncio.sleep(0.1)

    async def _process_batch(self, batch: List[Dict]) -> List[Dict]:
        raise NotImplementedError



from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field


class ServiceConfig(BaseModel):
    name: str
    engine: str
    # collection: str # mostly for book keeping purpose and allow service name to be cleaner
    config: Dict[str, Any]
    processor: str = "BatchQueryProcessor"
    cache: int = -1
    batch_size: int = 32
    cache_ttl: int = 600
    max_wait_time: float = 0.05
    cache_key_fields: List[str] = Field(default_factory=lambda: ["query", "limit"])
    cache_redis_url: Optional[str] = None
    cache_redis_kwargs: Optional[Dict[str, Any]] = Field(default_factory=lambda: {})
    scoring_disabled: bool = False


class ColllectionConfig(BaseModel):
    name: str
    doc_path: str
    offset_source: Literal["msmarco_seg", "offsetfile"] = "offsetfile"
    id_field: str = "id"
    content_field: Union[str, List[str]] = "text"
    id_to_lang_mapping: Optional[str] = None
    cache_path: Optional[str] = None

    def model_post_init(self, __context):
        if not isinstance(self.content_field, list):
            self.content_field = [self.content_field]


class Config(BaseModel):
    services: List[ServiceConfig] = Field(default_factory=list)
    collections: List[ColllectionConfig] = Field(default_factory=list)
    server_imports: List[str] = Field(default_factory=list)  # not yet implemented
    file_imports: List[str] = Field(default_factory=list)
    dynamic_pipeline: bool = True



import time
from typing import Any, Dict, List

from ..models import Engine
from .abstract import BatchProcessor, Processor


class AsyncQueryProcessor(Processor):
    """Processor that serve every request independently through async calls.
    Should be used when all the engine does is issuing async calls
    """

    def __init__(self, engine: Engine, cache_size=1024, cache_ttl=600, cache_key=None, **kwargs):
        super().__init__(cache_size, cache_ttl, cache_key)
        self.engine = engine

    async def _submit(self, item: Dict[str, Any]):
        """
        Process one single item
        """

        query = item.pop("query")
        limit = int(item.pop("limit", 10))
        # TODO: This is getting increasingly annoying... should just use subset instead of subsets
        subsets = item.pop("subset", None)

        ranking = (await self.engine.search_batch([query], limit=[limit], subsets=[subsets]))[0]

        # Process each item in the batch
        return {"query": query, "scores": ranking, "service": self.engine.name, "processed": True, "timestamp": time.time()}


class BatchQueryProcessor(BatchProcessor):
    def __init__(self, engine: Engine, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine

    async def _process_batch(self, batch: List[Dict]) -> List[Dict]:
        """
        Process a batch of items.
        Override this method for specific batch processing logic.
        """
        # Simulate processing time (e.g., model inference)
        # await asyncio.sleep(0.5)
        queries = [item.get("query", "") for item in batch]
        limits = [int(item.get("limit", 10)) for item in batch]
        subsets = [item.get("subset", None) for item in batch]
        rankings = await self.engine.search_batch(queries, limit=limits, subsets=subsets)

        # Process each item in the batch
        results = []
        for item, ranking in zip(batch, rankings):
            # Example processing - just compute text length
            query = item.get("query", "")

            result = {"query": query, "scores": ranking, "service": self.engine.name, "processed": True, "timestamp": time.time()}
            results.append(result)

        return results



import asyncio
from pathlib import Path
from typing import List

import aiohttp

from ..models import Engine, Relay
from ..processors import (
    AsyncPairwiseScoreProcessor,
    AsyncQueryProcessor,
    BatchPairwiseScoreProcessor,
    ContentProcessor,
    Processor,
    ProcessorRegistry,
)
from ..utils import logger, session_request
from ..utils.extensions import load_all_extensions
from .config import Config


async def auto_add_relay_services(servers: List[str]):
    if isinstance(servers, str):
        servers = [servers]

    async with aiohttp.ClientSession() as session:
        resps = await asyncio.gather(
            *[session_request(session, url=f"{server}/avail", method="GET") for server in servers]
        )

    # ensure backward compatible
    avail_services = {
        server: {
            "search": resp['search'] if 'search' in resp else resp['query'],
            "score": resp['score']
        }
        for server, resp in zip(servers, resps)
    }

    for server in servers:
        for service_type, processor_cls in zip(["search", "score"], [AsyncQueryProcessor, AsyncPairwiseScoreProcessor]):
            for service_name in avail_services[server][service_type]:
                if ProcessorRegistry.has_service(service_name, service_type):
                    continue
                logger.info(f"Adding auto Relay to {server} for service `{service_name}` of type {service_type}")
                processor = processor_cls(
                    engine=Relay(name=service_name, config={"endpoint": server, "service": service_name})
                )
                await processor.start()
                ProcessorRegistry.register(service_name, service_type, processor)


async def load_config(config: str):
    if Path(config).exists():
        config = Path(config).read_text()

    config: Config = Config.model_validate_json(config)

    load_all_extensions(user_specified_files=config.file_imports)

    for collection_config in config.collections:
        ProcessorRegistry.register(collection_config.name, "content", ContentProcessor(collection_config))
    logger.info("All collections are loaded")

    for service_config in config.services:
        def _cache_key(x):
            return tuple(x.get(k, "") for k in service_config.cache_key_fields)

        engine: Engine = Engine.load(service_config.engine, name=service_config.name, config=service_config.config)

        if engine.can_search:
            processor: Processor = Processor.load(
                service_config.processor,
                engine=engine,
                batch_size=service_config.batch_size,
                max_wait_time=service_config.max_wait_time,
                cache_size=service_config.cache,
                cache_ttl=service_config.cache_ttl,
                cache_key=_cache_key,
                redis_url=service_config.cache_redis_url,
                redis_kwargs=service_config.cache_redis_kwargs,
            )
            await processor.start()
            ProcessorRegistry.register(service_config.name, "search", processor)

        if engine.can_score and not service_config.scoring_disabled:
            processor = BatchPairwiseScoreProcessor(
                engine,
                batch_size=service_config.batch_size,
                max_wait_time=service_config.max_wait_time,
                cache_size=-1,  # turn off cache for now
            )
            await processor.start()
            ProcessorRegistry.register(service_config.name, "score", processor)

        logger.info(f"{service_config.name} initialized and ready")

    await auto_add_relay_services(config.server_imports)

    logger.info("All services are initialized")



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Union

from ..utils import load_singleton, logger
from .abstract import Engine


try:
    from bsparse.anserini import Anserini
    from bsparse.models import SpladeModel
except ImportError:
    logger.warning("Failed to import bsparse for LSR")


class LSR(Engine):
    def __init__(self, name: str = "LSR", config: Union[str, Path, Dict[str, Any]] = None, **kwargs):
        super().__init__(name, config, **kwargs)

        index = self.config["index_path"]
        self.anserini = Anserini(index)

        @dataclass
        class Args:
            model_name: str = self.config["model_name"]
            max_length: int = self.config["max_length"]
            batch_size: int = self.config["batch_size"]
            max_terms: int = self.config["max_terms"]
            tokenizer_name: str = None

        config = Args()
        # we use SpladeV3 as the query encoder when MultiLSR is the document encoder
        self.model = SpladeModel(config)

        self.subset_mapper: Dict[str, str] = None
        if "id_to_subset_mapping" in self.config:
            if self.config["id_to_subset_mapping"].endswith(".pkl"):
                self.subset_mapper = load_singleton(self.config["id_to_subset_mapping"])
            else:
                logger.warning(f"Unable to load subset mapping file {self.config['id_to_subset_mapping']}")

    def filter_subset(self, scores: Dict[str, float], only_subset: str = None):
        if only_subset is None or self.subset_mapper is None:
            return scores
        return {doc_id: score for doc_id, score in scores.items() if self.subset_mapper[doc_id] == only_subset}

    async def search_batch(
        self, queries: List[str], limit: Union[int, List[int]] = 20, subsets: List[str] = None, maxp: bool = True
    ) -> List[Dict[str, float]]:
        if isinstance(limit, int):
            limit = [int(limit)] * len(queries)

        if subsets is None:
            subsets = [None] * len(queries)

        # anserini only takes 1 k, send max and remove any extra results.
        results = self.anserini.query_from_raw_text(queries, model=self.model, k=max(limit) * self.config.get("k_scale", 20))
        return [
            dict(sorted(self.filter_subset(scores, subset).items(), key=lambda x: -x[1])[:l])
            for subset, l, scores in zip(subsets, limit, results)
        ]



from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

from ..utils import dict_topk, load_singleton, logger, pbar
from .abstract import Aggregation, Engine


try:
    import torch
    from colbert import Searcher
    from colbert.infra import ColBERTConfig, Run, RunConfig
    from colbert.modeling.checkpoint import Checkpoint
    from colbert.modeling.colbert import ColBERT
    from colbert.search.index_storage import IndexScorer
except ImportError:
    logger.warning("Failed to import packages for PLAID-X")

_plaidx_checkpoint_singletons: Dict[str, Tuple["Checkpoint", int]] = {}


def colbert_all_pair_scores(Q: "torch.Tensor", D: "torch.Tensor", Dm: "torch.Tensor" = None):
    import torch

    assert len(Q.shape) == 3
    assert len(D.shape) == 3
    assert D.shape[-1] == Q.shape[-1]
    if Dm is not None:
        assert Dm.shape[1] == D.shape[1]
    dim = Q.shape[-1]

    x = (Q.to(D.dtype).view(-1, dim) @ D.view(-1, dim).T).view(-1, *D.shape[:2])
    if Dm is not None:
        x = torch.mul(x, Dm.squeeze(2).unsqueeze(0))
    return x.max(-1).values.view(*Q.shape[:2], -1).sum(1)


def _load_mapping(fn, containing_passage_id=True):
    if containing_passage_id:
        return ["_".join(line.strip().split("\t")[1].split("_")[:-1]) for line in pbar(open(fn))]
    else:
        return [line.strip().split("\t")[1] for line in pbar(open(fn))]


def _cumsum(arr):
    return [sum(arr[: i + 1]) for i in range(len(arr))]


class PLAIDX(Engine):
    def __init__(self, name: str = "PLAID-X", config: Union[str, Path, Dict[str, Any]] = None, **kwargs):
        super().__init__(name, config, **kwargs)

        # make sure ninja can load
        IndexScorer.try_load_torch_extensions(False)

        self.colbert_config = ColBERTConfig.load_from_index(self.index_path)
        if "checkpoint" in self.config:
            self.colbert_config.checkpoint = self.config["checkpoint"]

        checkpoint = PLAIDX._get_existing_checkpoint_instance(self.colbert_config.checkpoint)

        if checkpoint is not None:
            logger.info("===== Reuse model!")
        else:
            ColBERT.try_load_torch_extensions(False)  # make sure it is loaded
            checkpoint = Checkpoint(self.colbert_config.checkpoint, colbert_config=self.colbert_config)
            if config.get("use_gpu", False):
                checkpoint = checkpoint.half().to(device=f"cuda:{self.config.get('gpu_assignment', 0)}")
                checkpoint.compile()
            PLAIDX._cache_checkpoint_instance(self.colbert_config.checkpoint, checkpoint)

        # use_gpu = int(config.get('use_gpu', False)))

        self.passage_mapper = None
        if "passage_mapping" in self.config:
            self.passage_mapper = Aggregation(
                _load_mapping(self.config["passage_mapping"], self.config.get("mapping_containing_passage_id", True))
            )

        self.subset_mapper: Dict[str, str] = None
        if "id_to_subset_mapping" in self.config:
            if self.config["id_to_subset_mapping"].endswith(".pkl"):
                self.subset_mapper = load_singleton(self.config["id_to_subset_mapping"])
            else:
                logger.warning(f"Unable to load subset mapping file {self.config['id_to_subset_mapping']}")

        with Run().context(RunConfig(index_root=self.index_path.parent, gpus=0)):
            logger.info(f"Loading PLAID-X index `{self.index_path}` and checkpoint `{self.colbert_config.checkpoint}`")
            self.searcher = Searcher(index=self.index_path.name, checkpoint=checkpoint, load_collection=False)

        self.inference_batch_size = int(config.get("inference_batch_size", 32))

    def __del__(self):
        if hasattr(self, "searcher"):
            PLAIDX._delete_checkpoint_reference(self.searcher.checkpoint)

    @staticmethod
    def _get_existing_checkpoint_instance(checkpoint_path: str):
        global _plaidx_checkpoint_singletons
        abs_path = Path(checkpoint_path).absolute()
        if abs_path not in _plaidx_checkpoint_singletons:
            return None
        else:
            _plaidx_checkpoint_singletons[abs_path] = (
                _plaidx_checkpoint_singletons[abs_path][0],
                _plaidx_checkpoint_singletons[abs_path][1] + 1,
            )
            return _plaidx_checkpoint_singletons[abs_path][0]

    @staticmethod
    def _cache_checkpoint_instance(checkpoint_path: str, checkpoint: "Checkpoint"):
        global _plaidx_checkpoint_singletons
        abs_path = Path(checkpoint_path).absolute()
        assert abs_path not in _plaidx_checkpoint_singletons
        _plaidx_checkpoint_singletons[abs_path] = (checkpoint, 1)

    @staticmethod
    def _delete_checkpoint_reference(checkpoint: "Checkpoint"):
        global _plaidx_checkpoint_singletons
        for path in _plaidx_checkpoint_singletons:
            if _plaidx_checkpoint_singletons[path][0] == checkpoint:
                _plaidx_checkpoint_singletons[path] = (
                    _plaidx_checkpoint_singletons[path][0],
                    _plaidx_checkpoint_singletons[path][1] - 1,
                )
                if _plaidx_checkpoint_singletons[path][1] == 0:
                    del _plaidx_checkpoint_singletons[path]
                return

    def filter_subset(self, scores: Dict[str, float], only_subset: str = None):
        if only_subset is None or self.subset_mapper is None:
            return scores
        return {doc_id: score for doc_id, score in scores.items() if self.subset_mapper[doc_id] == only_subset}

    async def search_batch(
        self, queries: List[str], limit: Union[int, List[int]] = 20, subsets: List[str] = None, maxp: bool = True
    ) -> List[Dict[str, float]]:
        if isinstance(limit, int):
            limit = [int(limit * 1.5)] * len(queries)

        if subsets is None:
            subsets = [None] * len(queries)

        Q = self.searcher.encode(queries)
        scores = []

        for query_idx, (k, sub) in enumerate(zip(limit, subsets)):
            # TODO: expose these settings into config
            if k <= 10:
                self.searcher.configure(ncells=1, centroid_score_threshold=0.5, ndocs=256)
            elif k <= 100:
                self.searcher.configure(ncells=2, centroid_score_threshold=0.45, ndocs=1024)
            else:
                self.searcher.configure(ncells=2, centroid_score_threshold=0.4, ndocs=2048)

            pids, q_scores = self.searcher.ranker.rank(self.searcher.config, Q[query_idx : query_idx + 1])
            scores.append(dict(zip(pids, q_scores)))
            # s = dict((pid, score) for pid, _ , score in zip(*self.searcher.dense_search(Q[query_idx:query_idx+1], max(100, k*2))) )

        if not maxp or self.passage_mapper is None:
            return scores

        return [
            dict_topk(self.filter_subset(scores, subset), l)
            for subset, l, scores in zip(subsets, limit, map(self.passage_mapper.maxp, scores))
        ]

    async def score_batch(self, queries: List[str], passages: List[str], candidate_length: List[int]) -> List[List[float]]:
        assert len(candidate_length) == len(queries)
        assert sum(candidate_length) == len(passages)
        offsets = _cumsum([0] + candidate_length)

        import torch

        with torch.inference_mode():
            Q = self.searcher.checkpoint.queryFromText(queries)
            D, Dm = self.searcher.checkpoint.doc(
                *self.searcher.checkpoint.doc_tokenizer.tensorize(passages), keep_dims="return_mask"
            )

            return [
                colbert_all_pair_scores(Q[[i]], D[bidx:eidx], Dm[bidx:eidx]).cpu().tolist()[0]
                for i, (bidx, eidx) in enumerate(zip(offsets[:-1], offsets[1:]))
            ]



import io
import json
import pickle
from functools import partial
from pathlib import Path
from typing import Callable, Dict

from ..utils import logger, pbar


class RandomAccessReader:
    def __init__(self, path: Path):
        self.path = Path(path)

    def __getitem__(self, idx: str) -> str:
        raise NotImplementedError

    def __contains__(self, idx: str) -> bool:
        raise NotImplementedError


class OffsetFile(RandomAccessReader):
    def __init__(self, path: Path, key: Callable = None, offset_fn: Path = None, id_field: str = "id"):
        """
        Initialize OffsetFile.

        Args:
            fn: Path to the document file
            key: Callable to extract key from line (legacy support)
            offset_fn: Path to store the offset map
            num_workers: Number of parallel workers
            id_field: Field name to use as key (preferred over key function)
        """
        super().__init__(path)

        if offset_fn is None:
            offset_fn = self.path.parent / (self.path.name + ".offsetmap")
        else:
            offset_fn = Path(offset_fn)

        # Prefer id_field over key function for parallel processing
        self.id_field = id_field
        self.key_func = key

        if not offset_fn.exists():
            logger.info(f"Building offset map for {self.path}...")
            self.create_offsetmap(self.path, offset_fn, key or self._default_key)
        else:
            logger.info(f"Loading existing offset map from {offset_fn}")

        try:
            loaded_fn, self.pointer_dict = pickle.load(offset_fn.open("rb"))
            logger.info(f"Loaded offset map with {len(self.pointer_dict):,} entries")
        except Exception as e:
            logger.error(f"Failed to load offset map from {offset_fn}: {e}")
            raise

        self.fread = self.path.open("rt")

    def _default_key(self, line: str) -> str:
        """Default key extraction using id_field."""
        data = json.loads(line)
        return str(data[self.id_field]).strip()

    def create_offsetmap(self, fn: Path, offset_fn: Path, key: Callable[[str], str]):
        """Fallback sequential implementation with optimizations."""
        mapping = {}

        # Use larger buffer size for reading
        buffer_size = 1024 * 1024  # 1MB buffer

        with fn.open("rt", buffering=buffer_size) as fr:
            count_err = 0
            line_count = 0

            # Update progress less frequently
            update_frequency = 10000

            with pbar(desc=f"Building offset map for {fn} (sequential)") as progress:
                while True:
                    loc = fr.tell()
                    line = fr.readline()
                    if line == "":
                        break

                    try:
                        mapping[key(line).strip()] = loc
                        count_err = 0
                    except Exception as e:
                        logger.warning(f"Offset #{loc} decode error: {e}")
                        count_err += 1

                    if count_err > 10:
                        raise Exception("Too many errors")

                    line_count += 1
                    if line_count % update_frequency == 0:
                        progress.update(update_frequency)

                # Update remaining lines
                progress.update(line_count % update_frequency)

        logger.info(f"Writing offset map to {offset_fn}")
        with open(offset_fn, "wb") as fw:
            pickle.dump((str(fn), mapping), fw, protocol=pickle.HIGHEST_PROTOCOL)
            fw.flush()

    def __getitem__(self, idx: str):
        if idx not in self.pointer_dict:
            return {}
        self.fread.seek(self.pointer_dict[idx])
        return self.fread.readline()

    def __contains__(self, idx: str):
        return idx in self.pointer_dict


class MSMARCOSegOffset(RandomAccessReader):
    def __init__(
        self, path: Path, num_workers: int = 8, filename_pattern="msmarco_v2.1_doc_segmented_{shard}.json.gz", id_parser=None
    ):
        super().__init__(path)

        try:
            from rapidgzip import RapidgzipFile

            self.opener = partial(RapidgzipFile, parallelization=num_workers)
        except:
            logger.warning("Failed loading rapidgzip for .gz collection. Failling back to native gzip, which is slower.")
            import gzip

            self.opener = gzip.open

        self.filename_pattern = filename_pattern
        self.id_parser = id_parser if id_parser is not None else self._parse_idx

        self.cached_fps: Dict[str, io.BytesIO] = {}

    def _parse_idx(self, idx: str):
        idx = idx.split("_")
        return idx[3], int(idx[5])

    def __getitem__(self, idx: str):
        shard, off = self.id_parser(idx)
        fn = str(self.path / self.filename_pattern.format(shard=shard))
        if fn not in self.cached_fps:
            self.cached_fps[fn] = self.opener(fn)
        fp = self.cached_fps[fn]
        fp.seek(off)
        return fp.readline().decode()

    def __contains__(self, idx: str):
        return self[idx] != ""



import asyncio
import os
import uuid
from functools import partial
from typing import Any, List

from ...utils import logger
from ..abstract import Reranker
from .prompts import PromptBuilder, ResponseParser


# engine = AsyncLLMEngine.from_engine_args(AsyncEngineArgs(model="Qwen/QwQ-32B", tensor_parallel_size=4, distributed_executor_backend="ray"))


# TODO: at some point thes should be class and implement the same interface and
# fold in the init in them, but probably good enough for now.
async def _openai_completion(client, model, prompts, **kwargs):
    if "temperature" not in kwargs:
        kwargs["temperature"] = 0.6
    if "max_tokens" not in kwargs:
        kwargs["max_tokens"] = 2000

    res = await asyncio.gather(
        *[
            client.chat.completions.create(model=model, messages=[{"role": "user", "content": prompt}], **kwargs)
            for prompt in prompts
        ]
    )

    return [r.choices[0].message.content.strip() for r in res]


# TODO: still buggy
async def _vllm_async_completion(engine, prompts, **kwargs):
    from vllm import SamplingParams

    # Submit all requests
    request_ids = [uuid.uuid4() for _ in range(len(prompts))]
    sampling_params = SamplingParams(*kwargs)

    # Add request to engine
    await asyncio.gather(
        *[
            engine.add_request(request_id=request_id, prompt=prompt, params=sampling_params)
            for request_id, prompt in zip(request_ids, prompts)
        ]
    )

    # Process results as they complete
    results = {}
    while len(results) < len(prompts):
        request_outputs = await engine.step_async()

        for request_output in request_outputs:
            if request_output.finished:
                results[request_output.request_id] = request_output.outputs[0].text.strip()

    return [results[request_id] for request_id in request_ids]


class LLMReranker(Reranker):
    def __init__(self, name=None, config=None, **kwargs):
        super().__init__(name, config, **kwargs)

        assert "model_name_or_path" in self.config

        if "server_endpoint" in self.config:
            import openai

            client = openai.AsyncOpenAI(
                api_key=self.config.get("api_key", os.getenv("OPENAI_API_KEY", "noneset")),
                base_url=self.config["server_endpoint"],
            )
            self.llm_caller = partial(
                _openai_completion, client, self.config["model_name_or_path"], **self.config.get("request_kws", {})
            )
        else:
            import torch
            from vllm import AsyncEngineArgs, AsyncLLMEngine

            engine = AsyncLLMEngine.from_engine_args(
                AsyncEngineArgs(
                    model=self.config["model_name_or_path"],
                    tensor_parallel_size=torch.cuda.device_count(),
                    distributed_executor_backend="ray",
                    **self.config.get("init_kws", {}),
                )
            )
            self.llm_caller = partial(_vllm_async_completion, engine, **self.config.get("request_kws", {}))

        self.truncator = lambda x: x
        if "truncate_doc_to_n_tokens" in self.config and self.config["truncate_doc_to_n_tokens"] > 0:
            from transformers import AutoTokenizer

            tokenizer = AutoTokenizer.from_pretrained(
                self.config.get("tokenizer_name", self.config["model_name_or_path"]), use_fast=True
            )
            self.truncator = lambda x: tokenizer.decode(tokenizer.encode(x)[: self.config["truncate_doc_to_n_tokens"]])

        self.prompt_builder = PromptBuilder.load(self.config.get("prompt_builder", "RankKPromptBuilder"))
        self.response_parser = ResponseParser.load(self.config.get("response_parser", "NumberListParser"))

        self.window_size = int(self.config.get("rerank_window", 20))
        self.stride_size = int(self.config.get("rerank_stride", 10))

    async def rerank_call(self, query, doc_ids, candidates):
        resp = await self.llm_caller(self.prompt_builder(query, candidates))
        new_ranking: List[Any] = self.response_parser(resp, doc_ids)

        # new_ranking = [
        #     doc_id for doc_id, score in sorted(scores.items(), key=lambda x: -x[1])
        # ]

        if len(new_ranking) < len(doc_ids):
            adding = [d for d in doc_ids if d not in new_ranking]
            print(f"{new_ranking} ------ {adding}")
            new_ranking += [d for d in doc_ids if d not in new_ranking]

        assert len(doc_ids) == len(new_ranking)
        # , f"========== {len(doc_ids)}, {len(new_ranking)}, {new_ranking}"

        return new_ranking

    async def rerank_single_query(self, query: str, docs: List[str], stride_size: int = None, window_size: int = None):
        global_doc_idx = list(range(len(docs)))
        stride_size = stride_size if stride_size is not None else self.stride_size
        window_size = window_size if window_size is not None else self.window_size

        for rerank_end_idx in range(len(docs), 0, -stride_size):
            rerank_start_idx = max(rerank_end_idx - window_size, 0)

            logger.info(f"Reranking idx {rerank_start_idx} to {rerank_end_idx}")
            doc_ids = global_doc_idx[rerank_start_idx:rerank_end_idx].copy()

            global_doc_idx[rerank_start_idx:rerank_end_idx] = await self.rerank_call(
                query, doc_ids, [docs[did] for did in doc_ids]
            )

            if rerank_start_idx == 0:
                break

        return global_doc_idx

    async def score(self, queries, all_docs, candidate_length=None, with_progress=False, **kwargs):
        # TODO: worth later on make it single call instead of batching to support async
        # Batching is probably a bad idea here.
        truncated_docs = list(map(self.truncator, all_docs))

        if candidate_length is None:
            assert len(queries) == 1
            candidate_length = [len(all_docs)]

        s = 0
        tasks = []
        for query, n_docs in zip(queries, candidate_length):
            tasks.append(
                self.rerank_single_query(
                    query=query,
                    docs=truncated_docs[s : s + n_docs],
                    stride_size=kwargs.get("stride_size"),
                    window_size=kwargs.get("window_size"),
                )
            )
            s += n_docs

        results = await asyncio.gather(*tasks)

        s = 0
        all_scores = {}
        for qidx, (reranked_list, n_docs) in enumerate(zip(results, candidate_length)):
            assert len(reranked_list) == n_docs
            for pos, local_didx in enumerate(reranked_list):
                all_scores[(qidx, s + local_didx)] = 1 / (pos + 1)
            s += n_docs

        return all_scores



from typing import Any, Dict, List

from .abstract import Aggregation, Engine, Reranker
from .fusion import Fusion
from .lsr import LSR
from .mt5 import MT5Reranker
from .plaidx import PLAIDX
from .qwen3 import Qwen3
from .relay import Relay



import os
from pathlib import Path
from typing import Any, Dict, List, Union

import numpy as np
import openai
from trecrun import TRECRun

from ..utils import dict_topk, load_singleton, logger
from .abstract import Engine


try:
    import faiss
except ImportError:
    logger.warning("Failed to import Faiss for Qwen3")


class Qwen3(Engine):
    def __init__(self, name: str = "Qwen3", config: Union[str, Path, Dict[str, Any]] = None, **kwargs):
        super().__init__(name, config, **kwargs)

        self.client = openai.AsyncOpenAI(
            api_key=self.config.get("api_key", os.getenv("OPENAI_API_KEY", "noneset")), base_url=self.config["embedding_base_url"]
        )
        self.embedding_model_name = self.config.get("embedding_model_name", "Qwen/Qwen3-Embedding-8B")

        index_dir = Path(self.config["index_path"])
        index_path = index_dir / "index.faiss"
        ids_path = index_dir / "index.ids"

        logger.info(f"Loading FAISS index from: {index_path}")
        self.index = faiss.read_index(str(index_path))

        logger.info(f"Loading document IDs from: {ids_path}")
        with ids_path.open("r") as f:
            self.doc_ids = [line.strip() for line in f]

        logger.info(f"Index contains {self.index.ntotal} vectors")

        self.subset_mapper: Dict[str, str] = None
        if "id_to_subset_mapping" in self.config:
            if self.config["id_to_subset_mapping"].endswith(".pkl"):
                self.subset_mapper = load_singleton(self.config["id_to_subset_mapping"])
            else:
                logger.warning(f"Unable to load subset mapping file {self.config['id_to_subset_mapping']}")

    def filter_subset(self, scores: Dict[str, float], only_subset: str = None):
        if only_subset is None or self.subset_mapper is None:
            return scores
        return {doc_id: score for doc_id, score in scores.items() if self.subset_mapper[doc_id] == only_subset}

    def add_query_instructions(self, queries) -> List[str]:
        # TODO: move this to the config
        task_description = "Given a web search query, retrieve relevant passages that answer the query"
        return [f"Instruct: {task_description}\nQuery:{query}" for query in queries]

    async def search_batch(
        self, queries: List[str], limit: Union[int, List[int]] = 20, subsets: List[str] = None, maxp: bool = True
    ) -> List[Dict[str, float]]:
        if isinstance(limit, int):
            limit = [int(limit)] * len(queries)

        if subsets is None:
            subsets = [None] * len(queries)

        instruct_queries = self.add_query_instructions(queries)
        query_embeddings = await self.client.embeddings.create(model=self.embedding_model_name, input=instruct_queries)
        query_embeddings = np.array([x.embedding for x in query_embeddings.data])
        logger.debug(query_embeddings.shape)
        scores, ids = self.index.search(x=query_embeddings, k=max(limit) * self.config.get("k_scale", 20))

        qmap = dict(enumerate(queries))
        run = TRECRun({qid: dict(zip([self.doc_ids[x] for x in ids[qid]], scores[qid])) for qid in qmap})
        results = [run[str(qid)] for qid, _ in enumerate(queries)]

        return [dict_topk(self.filter_subset(scores, subset), l) for subset, l, scores in zip(subsets, limit, results)]



import json
import time
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Optional

from ..utils import logger


try:
    import redis.asyncio as redis
except ImportError:
    logger.warning("Redis is not installed, can only support in-memory caching.")


class Cache(ABC):
    """Abstract base class for cache implementations."""

    def __init__(self, capacity: int = -1, ttl: float = -1):
        """
        Initialize cache.

        Args:
            capacity: Maximum number of items to store (-1 for unlimited)
            ttl: Time-to-live in seconds for cache entries (-1 for no expiry)
        """
        self.capacity = capacity
        self.ttl = ttl

    @abstractmethod
    async def get(self, key: str) -> Optional[Any]:
        """
        Get an item from the cache.

        Args:
            key: The cache key

        Returns:
            The cached value or None if not found or expired
        """
        pass

    @abstractmethod
    async def put(self, key: str, value: Any) -> None:
        """
        Add an item to the cache.

        Args:
            key: The cache key
            value: The value to cache
        """
        pass

    @abstractmethod
    async def remove(self, key: str) -> None:
        """
        Remove an item from the cache.

        Args:
            key: The cache key
        """
        pass

    @abstractmethod
    async def clear(self) -> None:
        """Clear all items from the cache."""
        pass

    @abstractmethod
    async def __len__(self) -> int:
        """Return the number of items in the cache."""
        pass

    async def connect(self) -> None:
        """Connect to cache backend (if needed). Override in subclasses."""
        pass

    async def close(self) -> None:
        """Close cache connection (if needed). Override in subclasses."""
        pass

    async def __aenter__(self):
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.close()


class LRUCache(Cache):
    """An async LRU cache implementation."""

    def __init__(self, capacity: int = -1, ttl: float = -1):
        """
        Initialize LRU cache.

        Args:
            capacity: Maximum number of items to store (-1 for unlimited)
            ttl: Time-to-live in seconds for cache entries (-1 for no expiry)
        """
        super().__init__(capacity, ttl)
        self.cache = OrderedDict()
        self.timestamps = {}

    async def get(self, key: str) -> Optional[Any]:
        """
        Get an item from the cache.

        Args:
            key: The cache key

        Returns:
            The cached value or None if not found or expired
        """
        if key not in self.cache:
            return None

        # Check if the item has expired
        if self.ttl > 0:
            timestamp = self.timestamps[key]
            if time.time() - timestamp > self.ttl:
                # Item has expired
                await self.remove(key)
                return None

        # Move the item to the end (most recently used)
        self.cache.move_to_end(key)
        return self.cache[key]

    async def put(self, key: str, value: Any) -> None:
        """
        Add an item to the cache.

        Args:
            key: The cache key
            value: The value to cache
        """
        # If the key exists, update it and move to end
        if key in self.cache:
            self.cache[key] = value
            self.cache.move_to_end(key)
            self.timestamps[key] = time.time()
            return

        # Add new item
        self.cache[key] = value
        self.timestamps[key] = time.time()

        # If over capacity, remove the oldest item
        if self.capacity > 0 and len(self.cache) > self.capacity:
            oldest_key, _ = self.cache.popitem(last=False)
            del self.timestamps[oldest_key]

    async def remove(self, key: str) -> None:
        """Remove an item from the cache."""
        if key in self.cache:
            del self.cache[key]
            del self.timestamps[key]

    async def clear(self) -> None:
        """Clear the cache."""
        self.cache.clear()
        self.timestamps.clear()

    async def __len__(self) -> int:
        """Return the number of items in the cache."""
        return len(self.cache)


class RedisCache(Cache):
    """A Redis-based cache implementation with async support."""

    def __init__(
        self,
        capacity: int = -1,
        ttl: float = -1,
        redis_url: str = "redis://localhost:6379",
        key_prefix: str = "routircache:",
        **redis_kwargs,
    ):
        """
        Initialize Redis cache.

        Args:
            capacity: Maximum number of items to store (-1 for unlimited, not enforced by Redis)
            ttl: Time-to-live in seconds for cache entries (-1 for no expiry)
            redis_url: Redis connection URL
            key_prefix: Prefix for all cache keys to avoid collisions
            **redis_kwargs: Additional arguments to pass to Redis client
        """
        super().__init__(capacity, ttl)
        self.redis_url = redis_url
        self.key_prefix = key_prefix
        self.redis_kwargs = redis_kwargs
        self.redis_client: Optional[redis.Redis] = None

    async def connect(self):
        """Establish connection to Redis."""
        if self.redis_client is None:
            self.redis_client = await redis.from_url(self.redis_url, encoding="utf-8", decode_responses=True, **self.redis_kwargs)

    async def close(self):
        """Close Redis connection."""
        if self.redis_client:
            await self.redis_client.close()
            self.redis_client = None

    def _make_key(self, key: str) -> str:
        """Create a prefixed key."""
        # Convert key to string if it's a tuple or other type
        if isinstance(key, (tuple, list)):
            key = json.dumps(key)
        return f"{self.key_prefix}{key}"

    async def get(self, key: str) -> Optional[Any]:
        """
        Get an item from the cache.

        Args:
            key: The cache key

        Returns:
            The cached value or None if not found or expired
        """
        await self.connect()

        redis_key = self._make_key(key)
        value = await self.redis_client.get(redis_key)

        if value is None:
            return None

        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return value

    async def put(self, key: str, value: Any) -> None:
        """
        Add an item to the cache.

        Args:
            key: The cache key
            value: The value to cache
        """
        await self.connect()

        redis_key = self._make_key(key)

        # Serialize the value
        try:
            serialized_value = json.dumps(value)
        except (TypeError, ValueError):
            # If value can't be JSON serialized, convert to string
            serialized_value = str(value)

        # Set the value with optional TTL
        if self.ttl and self.ttl > 0:
            await self.redis_client.setex(redis_key, int(self.ttl), serialized_value)
        else:
            await self.redis_client.set(redis_key, serialized_value)

    async def remove(self, key: str) -> None:
        """Remove an item from the cache."""
        await self.connect()

        redis_key = self._make_key(key)
        await self.redis_client.delete(redis_key)

    async def clear(self) -> None:
        """Clear all cache entries with the current prefix."""
        await self.connect()

        # Find all keys with our prefix and delete them
        pattern = f"{self.key_prefix}*"
        cursor = 0

        while True:
            cursor, keys = await self.redis_client.scan(cursor, match=pattern, count=100)
            if keys:
                await self.redis_client.delete(*keys)
            if cursor == 0:
                break

    async def __len__(self) -> int:
        """Return the number of items in the cache with the current prefix."""
        await self.connect()

        pattern = f"{self.key_prefix}*"
        count = 0
        cursor = 0

        while True:
            cursor, keys = await self.redis_client.scan(cursor, match=pattern, count=100)
            count += len(keys)
            if cursor == 0:
                break

        return count



import argparse
from pathlib import Path

import faiss
import numpy as np
from tqdm.auto import tqdm


faiss.omp_set_num_threads(32)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_dir",
        help="A directory with *.npy files, each containing a two varaibles: features and ids. "
        "`Ids` should be a list of document ids, and `features` should be a 2D matrix with "
        "the first dimenion matching the length of `ids`.",
    )
    parser.add_argument("output_dir", help="Output directory for the index")

    parser.add_argument("--index_string", default="PQ2048x4fs")
    parser.add_argument("--sampling_rate", type=float, default=0.07)

    parser.add_argument("--use_gpu", action="store_true", default=False)
    parser.add_argument("--two_step_training", action="store_true", default=False)

    args = parser.parse_args()

    output_dir = Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    all_fns = list(Path(args.input_dir).glob("*.npy"))

    sampled_fns = all_fns[:: int(1 / args.sampling_rate)]
    sampled_vectors = np.concatenate([np.load(fn, allow_pickle=True).item()["features"] for fn in tqdm(sampled_fns)], axis=0)

    # drop example with na
    sampled_vectors = sampled_vectors[~np.isnan(sampled_vectors).any(axis=1)]

    index = faiss.index_factory(sampled_vectors.shape[1], args.index_string, faiss.METRIC_INNER_PRODUCT)

    if args.use_gpu:
        if args.two_step_training:
            index_ivf = faiss.extract_index_ivf(index)
            clustering_index = faiss.index_cpu_to_all_gpus(faiss.IndexFlatIP(index_ivf.d))
            index_ivf.clustering_index = clustering_index
        else:
            co = faiss.GpuMultipleClonerOptions()
            co.allowCpuCoarseQuantizer = True
            index = faiss.index_cpu_to_all_gpus(index)

    print("training...")
    index.train(sampled_vectors)

    if args.use_gpu:
        index = faiss.index_gpu_to_cpu(index)

    docids = []
    for fn in tqdm(all_fns, desc="adding", dynamic_ncols=True):
        shard = np.load(fn, allow_pickle=True).item()
        # dropna features
        mask = ~np.isnan(shard["features"]).any(axis=1)
        features = shard["features"][mask]
        ids = np.array(shard["ids"])[mask].tolist()

        index.add(features)
        docids += ids

    print("saving faiss index")
    faiss.write_index(index, str(output_dir / "index.faiss"))

    print("saving doc ids")
    with (output_dir / "index.ids").open("w") as fw:
        for docid in docids:
            fw.write(f"{docid}\n")
